(* utility functions -------------------------------------------------------- *)

let update : type a b. (a -> b) -> a -> b -> (a -> b) =
  fun f a b a' -> if a = a' then b else f a'

let next_mul n m : int64 =
  Int64.(div n m |> succ |> mul m)

let ( +. ) = Int64.add
let ( -. ) = Int64.sub
let ( *. ) = Int64.mul

type int64' = Int64_overflow.t

(* assembler syntax --------------------------------------------------------- *)

type lbl = string

type quad = int64

type imm = Lit of quad
         | Lbl of lbl

(* arguments: rdi, rsi, rdx, rcx, r09, r08
   callee-save rbx, rbp, r12-r15 *)
type reg = Rip
         | Rax | Rbx | Rcx | Rdx | Rsi | Rdi | Rbp | Rsp
         | R08 | R09 | R10 | R11 | R12 | R13 | R14 | R15

type operand = Imm of imm            (* immediate *)
             | Reg of reg            (* register *)
             | Ind1 of imm           (* indirect: displacement *)
             | Ind2 of reg           (* indirect: (%reg) *)
             | Ind3 of (imm * reg)   (* indirect: displacement(%reg) *)

type opcode = Movq | Pushq | Popq
            | Leaq
            | Incq | Decq | Negq | Notq
            | Addq | Subq | Imulq | Xorq | Orq | Andq
            | Shlq | Sarq | Shrq
            | Cmpq
            | Jmp | Je | Jne | Jg | Jge | Jl | Jle
            | Callq | Retq

type ins = opcode * operand list              
              

type 'a labled = { lbl: lbl; global: bool; value: 'a }

type data = Asciz of string
          | Quad of imm

type asm = { text_asm : ins list labled list
           ; data_asm : data labled list
           }


(* pretty printing ----------------------------------------------------------- *)

let string_of_reg : reg -> string = function
  | Rip -> "%rip"
  | Rax -> "%rax" | Rbx -> "%rbx" | Rcx -> "%rcx" | Rdx -> "%rdx"
  | Rsi -> "%rsi" | Rdi -> "%rdi" | Rbp -> "%rbp" | Rsp -> "%rsp"
  | R08 -> "%r8"  | R09 -> "%r9"  | R10 -> "%r10" | R11 -> "%r11"
  | R12 -> "%r12" | R13 -> "%r13" | R14 -> "%r14" | R15 -> "%r15"

let string_of_lbl (l:lbl) : string = l

let string_of_imm : imm -> string = function
  | Lit i -> Int64.to_string i
  | Lbl l -> string_of_lbl l

let string_of_operand : operand -> string = function
  | Imm i -> "$" ^ string_of_imm i
  | Reg r -> string_of_reg r
  | Ind1 i -> string_of_imm i
  | Ind2 r -> "(" ^ string_of_reg r ^ ")"
  | Ind3 (i, r) -> string_of_imm i ^ "(" ^ string_of_reg r ^ ")"

let string_of_jmp_operand : operand -> string = function
  | Imm i -> string_of_imm i
  | Reg r -> string_of_reg r
  | Ind1 i -> "*" ^ string_of_imm i
  | Ind2 r -> "*" ^ "(" ^ string_of_reg r ^ ")"
  | Ind3 (i, r) -> "*" ^ string_of_imm i ^ "(" ^ string_of_reg r ^ ")"

let string_of_opcode : opcode -> string = function
  | Movq -> "movq" | Pushq -> "pushq" | Popq -> "popq"
  | Leaq -> "leaq"
  | Incq -> "incq" | Decq -> "decq" | Negq -> "negq" | Notq -> "notq"
  | Addq -> "addq" | Subq -> "subq" | Imulq -> "imulq"
  | Xorq -> "xorq" | Orq -> "orq"  | Andq -> "andq"
  | Shlq -> "shlq" | Sarq -> "sarq" | Shrq -> "shrq"
  | Cmpq -> "cmpq"
  | Jmp -> "jmp" | Je -> "je" | Jne -> "jne" | Jg -> "jg" 
  | Jge -> "jge" | Jl -> "jl" | Jle -> "jle"
  | Callq -> "callq" | Retq -> "retq"

let string_of_ins (op, args: ins) : string =
  "\t" ^ string_of_opcode op ^ "\t" ^ String.concat ", " @@ 
  match op with
  | Jmp | Je | Jne | Jg | Jge | Jl | Jle | Callq -> List.map string_of_jmp_operand args
  | _                                            -> List.map string_of_operand args

let string_of_labled (f:'a -> string) (l:'a labled) : string =
  (if l.global then "\t.globl\t" ^ string_of_lbl l.lbl ^ "\n" else "")
  ^ string_of_lbl l.lbl ^ ":\n"
  ^ f l.value

let string_of_data : data -> string = function
  | Asciz s -> "\t.asciz\t" ^ "\"" ^ s ^ "\""
  | Quad i -> "\t.quad\t" ^ string_of_imm i

let string_of_asm (a:asm) : string =
  let map_concat_n f a = String.concat "\n" @@ List.map f a in
  "\t.text\n"
  ^ map_concat_n (string_of_labled @@ map_concat_n string_of_ins) a.text_asm
  ^ "\n\t.data\n"
  ^ map_concat_n (string_of_labled string_of_data) a.data_asm

(* simulator ---------------------------------------------------------------- *)

let mem_size = 0x10000          (* 64k *)
let nregs = 17                  (* including Rip *)
let ins_size = 4L               (* assume we have a 4-byte encoding *)
let exit_addr = 0xfdeadL        (* halt when m.regs(%rip) = exit_addr *)

type sbyte = InsB0 of ins       (* 1st byte of an instruction *)
           | InsFrag            (* 2nd, 3rd, or 4th byte of an instruction *)
           | Byte of char       (* non-instruction byte *)

type mem = sbyte array

type flags = { mutable fo : bool
             ; mutable fs : bool
             ; mutable fz : bool
             }

type regs = int64 array

type mach = { flags : flags
            ; regs : regs
            ; mem : mem
            }

let rind : reg -> int = function
  | Rip -> 16
  | Rax -> 0  | Rbx -> 1  | Rcx -> 2  | Rdx -> 3
  | Rsi -> 4  | Rdi -> 5  | Rbp -> 6  | Rsp -> 7
  | R08 -> 8  | R09 -> 9  | R10 -> 10 | R11 -> 11
  | R12 -> 12 | R13 -> 13 | R14 -> 14 | R15 -> 15


(* helper functions for reading machine state *)

let sbytes_of_int64 (i:int64) : sbyte list =
  let open Char in 
  let open Int64 in
  List.map (fun n -> Byte (shift_right i n |> logand 0xffL |> to_int |> chr))
           [0; 8; 16; 24]

let int64_of_sbytes (bs:sbyte list) : int64 =
  let open Char in
  let open Int64 in
  List.fold_right (fun b i -> match b with
                              | Byte c -> logor (shift_left i 8) (c |> code |> of_int)
                              | _ -> 0L) bs 0L

let sbytes_of_string (s:string) : sbyte list =
  let rec loop acc = function
    | i when i < 0 -> acc
    | i -> loop (Byte s.[i]::acc) (pred i)
  in
  loop [Byte '\x00'] @@ String.length s - 1

let readquad (m:mach) (addr:int64) : int64 =
  let base = Int64.to_int addr in
  int64_of_sbytes [m.mem.(base); m.mem.(base+1); m.mem.(base+2); m.mem.(base+3)]

let writequad (m:mach) (addr:int64) (w:int64) : unit =
  let base = Int64.to_int addr in
  Array.blit (Array.of_list @@ sbytes_of_int64 w) 0 m.mem base 4

let fetchins (m:mach) (addr:int64) : ins =
  match m.mem.(Int64.to_int addr) with
  | InsB0 i -> i
  | _ -> failwith @@ "fetchins: failed to decode instruction at " ^ Int64.to_string addr

(* It might be useful to toggle e.g. printing of intermediate states of your simulator. *)
let debug_simulator = ref false

let step (m:mach) : unit =
failwith "step unimplemented"

let run_until (addr:quad) (m:mach) : unit =
  while m.regs.(rind Rip) <> addr do step m done

(* Run the simulator until m.regs.(%rip) = exit_addr *)
let run (m:mach) : int64 = 
  run_until exit_addr m;
  m.regs.(rind Rax)

(* assembling and linking --------------------------------------------------- *)

(* A symbol table maps symbols to addresses *)
type symtab = lbl -> quad

(* A layout specifies the start address of each section *)
type layout = { text_pos: quad; data_pos: quad }

(* An object header contains metadata required for linking:
   - sec_sizes: the length of each section
   - symtab: a symbol table containing the *offsets* of each symbol from the
       beginning of its section *)
type objheader = { data_len : quad
                 ; text_len : quad
                 ; symtab   : layout -> symtab
                 }

(* An object file consists of a header and the assembled machine code, which
   consists of a list of symbolic bytes for each section (sec -> sbyte list).
   Since an object file can reference symbols in other compilation units, we
   represent the assembled code as a procedure for resolving symbols given
   a symbol table. *)
type objfile = { sym_offs : layout -> symtab
               ; data_len : quad
               ; text_len : quad
               ; data_sec : symtab -> sbyte list
               ; text_sec : symtab -> sbyte list
               } 

(* An executable file consists of code for each section (sec -> sbyte list), and
   some meta data required to *load* a program into memory and execute it.
   - entry: the address of the beginning of the program
   - seg_addr: the address that each section should be loaded into in memory *)
type execfile = { entry    : quad
                ; layout   : layout
                ; data_seg : sbyte list
                ; text_seg : sbyte list
                }

exception Unknown_sym of string

(* Given a symbol table, translate each section of a program to a list of bytes. *)
let assemble {text_asm; data_asm} : objfile =
  failwith "for HW2"

(* given any number of object files, create an executable file. *)
let link : objfile list -> execfile =
  fun x -> failwith "for HW2"

let load {entry; layout; data_seg; text_seg} : mach = 
  let module A = Array in

  let copy bs dest addr =
    A.blit (A.of_list bs) 0 dest (Int64.to_int addr) (List.length bs)
  in

  let mem = A.make mem_size (Byte '\x00') in
  copy text_seg mem layout.text_pos;
  copy data_seg mem layout.data_pos;

  let regs = A.make nregs 0L in
  regs.(rind Rip) <- entry;
  regs.(rind Rsp) <- Int64.of_int mem_size -. 8L;
  copy (sbytes_of_int64 exit_addr) mem regs.(rind Rsp);

  { flags = {fo = false; fs = false; fz = false}; regs; mem }

(* examples ------------------------------------------------------------------ *)

module Asm = struct
    let (~$) i = Imm (Lit (Int64.of_int i))
    let (~$$) l = Imm (Lbl l)
    let (~%) r = Reg r

    let gquad l q = { lbl = l; global = true; value = Quad q }
    let block l is = { lbl = l; global = false; value = is }
    let gblock l is = { lbl = l; global = true; value = is }
end

let p0 = Asm.{ text_asm =
                 [ block "foo"
                         [ Xorq, [~%Rax; ~%Rax]
                         ; Movq, [~$100; ~%Rax]
                         ; Retq, []
                         ]
                 ; block "main" 
                         [ Xorq, [~%Rax; ~%Rax]
                         ; Movq, [Ind1 (Lbl "baz"); ~%Rax]
                         ; Retq, []
                         ]
                 ]
             ; data_asm = 
                 [ { lbl = "baz"; global = true; value = Quad (Lit 99L) }
                 ; { lbl = "quux"; global = true; value = Asciz "Hello, world!" }
                 ]
             }

let p1 n = Asm.{ text_asm =
                   [ block "main"
                           [ Movq,  [~$1; ~%Rax]
                           ; Movq,  [~$n; ~%Rdi]
                           ]
                   ; block "loop"
                           [ Cmpq,  [~$0; ~%Rdi]
                           ; Je,    [~$$"exit"]
                           ; Imulq, [~%Rdi; ~%Rax]
                           ; Decq,  [~%Rdi]
                           ; Jmp,   [~$$"loop"]
                           ]
                   ; block "exit"
                           [ Retq,  [] 
                           ]
                   ]
               ; data_asm = []
               }

let p2 n = Asm.{ text_asm =
                   [ block "fac"
                           [ Movq,  [~%Rsi; ~%Rax]
                           ; Cmpq,  [~$1; ~%Rdi]
                           ; Je,    [~$$"exit"]
                           ; Imulq, [~%Rdi; ~%Rsi]
                           ; Decq,  [~%Rdi]
                           ; Callq, [~$$"fac"]
                           ]
                   ; block "exit"
                           [ Retq,  [] ]
                   ; gblock "_program"
                           [ Movq,  [~$n; ~%Rdi]
                           ; Movq,  [~$1; ~%Rsi]
                           ; Callq, [~$$"fac"]
                           ; Retq,  []
                           ]
                   ]
               ; data_asm = []
               }
